#include "ros/ros.h"
#include "std_msgs/String.h"

#include <sstream>

#include <SDL/SDL.h>
#include <termios.h>
#include "JS2Serial.h"
#include <fcntl.h>
#include <unistd.h>

#include "ros/time.h"
#include "sensor_msgs/Imu.h"
#include "MultiWiiSerial/mav_rcdata.h"
#include "MultiWiiSerial/mav_vbat.h"

using namespace std;

class Combiner{
public:
	Combiner(ros::NodeHandle node_handle_);
	~Combiner(){};

	void controlCallback(MultiWiiSerial::mav_rcdata msg);
private:
	ros::Publisher imu_data_pub_;
	ros::Publisher rx_data_pub_;
        ros::Publisher bat_data_pub_;
};

	//Node variables
	static int16_t sendData[8];        // interval [1000;2000]
	static int16_t rcData[8];          // interval [1000;2000]
	static float imu[9];
	static int8_t bytevbat;
	static int16_t pMeterSum;

	unsigned char inBuf[255];
	int offset = 0;
	int p;

	int serial;
        bool open_port_ok = false;
        int count_packet_send;


/*  ros::NodeHandle node_handle_;
  
  //Subscribers and publishers
  ros::Publisher imu_data_pub_;
  ros::Publisher rx_data_pub_;
  ros::Subscriber sub_move;*/

	sensor_msgs::Imu readingIMU;
	MultiWiiSerial::mav_rcdata readingRX;
        MultiWiiSerial::mav_vbat readingBat;

int16_t read16()
{
  int16_t s0, s1;
  s0 = inBuf[p++];
  s1 = inBuf[p++];
  
  return (int16_t)( (s0&0xff) + (s1<<8) );
  //return (int16_t)( s0 + s1 * 256 );  
}
int read8()
{ 
  int8_t s0;
  s0 = inBuf[p++];
  return (int8_t)(s0&0xff);
}


void evaluateCommand(char cmd)
{
  int i;
  int icmd = (int)(cmd);
  switch(icmd) {
    case MSP_IDENT:
        //version = read8();
        //multiType = read8();
        //read8(); // MSP version
        //multiCapability = read32();// capability
        break;
    case MSP_STATUS:
        //cycleTime = read16();
        //i2cError = read16();
        //present = read16();
        //mode = read32();
        break;
    case MSP_RAW_IMU:
        imu[0]   = read16();  imu[1] = read16();    imu[2] = read16();
        imu[3] = read16()/8;  imu[4] = read16()/8;  imu[5] = read16()/8;
        imu[6] = read16()/3;  imu[7] = read16()/3;  imu[8] = read16()/3;
	break;
    /*case MSP_SERVO:
        for(i=0;i<8;i++) servo[i] = read16(); break;
    case MSP_MOTOR:
        for(i=0;i<8;i++) mot[i] = read16(); break;*/
    case MSP_RC:
	//printf(">>>>>>>>");
	for(i=0;i<8;i++) {
	    rcData[i] = read16();
	    printf("%u ",rcData[i]);
	}
        printf("%u--\n" , icmd );
	break;
        //rcRoll = read16();rcPitch = read16();rcYaw = read16();rcThrottle = read16();    
        //rcAUX1 = read16();rcAUX2 = read16();rcAUX3 = read16();rcAUX4 = read16(); break;
    /*case MSP_RAW_GPS:
        GPS_fix = read8();
        GPS_numSat = read8();
        GPS_latitude = read32();
        GPS_longitude = read32();
        GPS_altitude = read16();
        GPS_speed = read16(); break;
    case MSP_COMP_GPS:
        GPS_distanceToHome = read16();
        GPS_directionToHome = read16();
        GPS_update = read8(); break;
    case MSP_ATTITUDE:
        angx = read16()/10;angy = read16()/10;
        head = read16(); break;
    case MSP_ALTITUDE:
        alt = read32(); break;*/
    case MSP_BAT:
        bytevbat = read8();
        pMeterSum = read16();
        break;
    /*case MSP_RC_TUNING:
        byteRC_RATE = read8();byteRC_EXPO = read8();byteRollPitchRate = read8();
        byteYawRate = read8();byteDynThrPID = read8();
        byteThrottle_MID = read8();byteThrottle_EXPO = read8();
        break;
    case MSP_ACC_CALIBRATION:
        break;
    case MSP_MAG_CALIBRATION:
        break;
    case MSP_PID:
        for(i=0;i<PIDITEMS;i++) {
          byteP[i] = read8();byteI[i] = read8();byteD[i] = read8();          
        }
        break;
    case MSP_BOX:
        for( i=0;i<CHECKBOXITEMS;i++) {
          activation[i] = read16();
        } break;
    case MSP_BOXNAMES:
        break;
    case MSP_PIDNAMES:
        break;
    case MSP_MISC:
        intPowerTrigger = read16();
        break;
    case MSP_MOTOR_PINS:
        for( i=0;i<8;i++) {
          byteMP[i] = read8();
        } break;
    case MSP_DEBUGMSG:
        //while(dataSize-- > 0) {
        //  char c = (char)read8();
        //  if (c != 0) {
        //    System.out.print( c );
        //  }
        //}
        break;
    case MSP_DEBUG:
        debug1 = read16();debug2 = read16();debug3 = read16();debug4 = read16(); break;
    default:
        //println("Don't know how to handle reply "+icmd);*/
  }
}


void sendMW(int cmd)
{
    uint8_t i;
    uint8_t z;
    uint8_t outputbuffer[256];
    uint8_t checksum=0; 
    int icmd = (int)(cmd);
	  
    switch(icmd) {
    case MSP_RAW_IMU:

	//header
	outputbuffer[0]='$';
	outputbuffer[1]='M';
	outputbuffer[2]='<';
	//size
	outputbuffer[3]=0;
	//message type
	outputbuffer[4]=MSP_RAW_IMU; //102;
	//CRC
	outputbuffer[5]=MSP_RAW_IMU; 


	 /*//for debug use
	 printf("send: ");
	 int j;
	 for(j=0;j<6;j++)
	 {
	 	printf("%d-",outputbuffer[j]);
	 }
	 printf("\n");*/
	write(serial,outputbuffer,6);
        count_packet_send++;
	break;
    case MSP_RC:

	//header
	outputbuffer[0]='$';
	outputbuffer[1]='M';
	outputbuffer[2]='<';
	//size
	outputbuffer[3]=0;
	//message type
	outputbuffer[4]=MSP_RC; //105;
	//CRC
	outputbuffer[5]=MSP_RC; 
	write(serial,outputbuffer,6);
        count_packet_send++;
	break;
    case MSP_SET_RAW_RC:
 
	checksum^=16;
	checksum^=200;
	//header
	outputbuffer[0]='$';
	outputbuffer[1]='M';
	outputbuffer[2]='<';
	//size
	outputbuffer[3]=16;
	//message type
	outputbuffer[4]=MSP_SET_RAW_RC; //200;
	
	z=0;
	for(i=5; i<21 ; i++)
	{
		//low byte
		outputbuffer[i]=(uint8_t)(sendData[z]&0xFF);
		checksum^=outputbuffer[i];
		i++;
		//high byte
		outputbuffer[i]=(uint8_t)(sendData[z]>>8);
		checksum^=outputbuffer[i];
	
		//printf("%u ",sendData[z]);
		z++;
	}
        //printf("%u--\n" , icmd );
	outputbuffer[21]=checksum;
	write(serial,outputbuffer,22);
        count_packet_send++;
	break;
    case MSP_BAT:

	//header
	outputbuffer[0]='$';
	outputbuffer[1]='M';
	outputbuffer[2]='<';
	//size
	outputbuffer[3]=0;
	//message type
	outputbuffer[4]=MSP_BAT; //110;
	//CRC
	outputbuffer[5]=MSP_BAT; 

	write(serial,outputbuffer,6);
        count_packet_send++;
	break;

    }
    //wait for serialport to send
    tcdrain(serial);
}

void getDataIMU(sensor_msgs::Imu& data)
{
//https://code.ros.org/svn/ros-pkg/stacks/imu_drivers/trunk/microstrain_3dmgx2_imu/imu_node.cc
    uint64_t time=0;

    data.linear_acceleration.x = imu[0];
    data.linear_acceleration.y = imu[1];
    data.linear_acceleration.z = imu[2];
 
    data.angular_velocity.x = imu[3];
    data.angular_velocity.y = imu[4];
    data.angular_velocity.z = imu[5];
      
    data.header.stamp = ros::Time::now().fromNSec(time);
}
void getDataRX(MultiWiiSerial::mav_rcdata& data)
{
    uint64_t time=0;
    uint8_t i;
    for(i=0;i<8;i++)
	{
	    data.channel[i] = rcData[i];
	}

    data.header.stamp = ros::Time::now().fromNSec(time);
}
void getDataBat(MultiWiiSerial::mav_vbat& data)
{
    uint64_t time=0;

    data.vBat = bytevbat;
    data.pMeterSum = pMeterSum;

    data.header.stamp = ros::Time::now().fromNSec(time);
}
void setSendDataRC(MultiWiiSerial::mav_rcdata& data)
{
    uint64_t time=0;
    uint8_t i;

    for(i=0;i<8;i++)
        {
	    sendData[i] = data.channel[i];
	    //printf("%d-",sendData[i]);
	}

	 //printf("\n");
}
void closeSerial()
{
    close (serial);
}

// ----------- Callback --------------------------------------------------------
void Combiner::controlCallback(MultiWiiSerial::mav_rcdata msg)
{
  uint8_t i;
  char buf[255];
  int res;
  char cmd;
  unsigned char c;
  int dataSize;
  char checksum;
  int err_rcvd=0;
  int c_state = IDLE;

  if (open_port_ok == false)  return;

  setSendDataRC(msg);

  sendMW(MSP_RAW_IMU);    // get IMU
  sendMW(MSP_RC);         // get RX
  sendMW(MSP_SET_RAW_RC); // set RC
  sendMW(MSP_BAT);        // set BAT

  int count = 0;
  //bool end_packet = false;
  while (count_packet_send > 0)
  {
		res = read(serial,buf,8);
		if(res>0) 
		{
		  for(i=0;i<res;i++)
		  {
		      c = (buf[i]&0xFF);
		      //printf("%x,%u_ ",c,c);
		      //printf("%u_ ",c);
  
		      if (c_state == IDLE) {
        		c_state = (c=='$') ? HEADER_START : IDLE;
		      } else if (c_state == HEADER_START) {
		        c_state = (c=='M') ? HEADER_M : IDLE;
		      } else if (c_state == HEADER_M) {
		        if (c == '>') {
		          c_state = HEADER_ARROW;
	              	} else if (c == '!') {
     			  c_state = HEADER_ERR;
		      	} else {
		          c_state = IDLE;
		      	}
		      } else if (c_state == HEADER_ARROW || c_state == HEADER_ERR) {
	     	        //is this an error message?
		        err_rcvd = (c_state == HEADER_ERR);        // now we are expecting the payload size
		        dataSize = (c&0xFF);
		        // reset index variables
		        p = 0;
		        offset = 0;
		        checksum = 0;
		        checksum ^= (c&0xFF);
		        // the command is to follow 
		        c_state = HEADER_SIZE;
		      } else if (c_state == HEADER_SIZE) {
		        cmd = (char)(c&0xFF);
		        checksum ^= (c&0xFF);
		        c_state = HEADER_CMD;
		      } else if (c_state == HEADER_CMD && offset < dataSize) {
		        checksum ^= (c&0xFF);
		        inBuf[offset++] = c;
		      } else if (c_state == HEADER_CMD && offset >= dataSize) {
		        // compare calculated and transferred checksum 
		        if ((checksum&0xFF) == (c&0xFF)) {
		          if (err_rcvd) {
				printf("Copter did not understand request type\n");
		          } else {
		            // we got a valid response packet, evaluate it 
		            evaluateCommand(cmd);
			    //end_packet = true;
			    count_packet_send = count_packet_send - 1;

			    ++count;
		          }
		        } else {
		        }
		        c_state = IDLE;
		
      		      } //if
   	 	  }//for
		}//if


  }//while

  

  //ROS_INFO("%s", "IMU sensor reading.");
 
  getDataIMU(readingIMU);
  readingIMU.header.frame_id = string("imu");
  imu_data_pub_.publish(readingIMU);
		    
  getDataRX(readingRX);
  readingRX.header.frame_id = string("RX");
  rx_data_pub_.publish(readingRX);

  getDataBat(readingBat);
  readingBat.header.frame_id = string("bat");
  bat_data_pub_.publish(readingBat);

  //ros::spinOnce(); ????
  //loop_rate.sleep(); ????
}

// ------------- Constructor -------------------------------------------------------------------
Combiner::Combiner(ros::NodeHandle node_handle_)
{

  this->imu_data_pub_ = node_handle_.advertise<sensor_msgs::Imu>("mwSerial/imu/data", 100);
  this->rx_data_pub_  = node_handle_.advertise<MultiWiiSerial::mav_rcdata>("mwSerial/rx/channel", 100);
  this->bat_data_pub_ = node_handle_.advertise<MultiWiiSerial::mav_vbat>("mwSerial/vBat/data", 100);

  //ros::Rate loop_rate(10);

  //open serialport
  serial = open ( SERIAL , O_RDWR | O_NOCTTY | O_NDELAY );
  if (serial == -1)
  {
    printf("error open port\n");
    open_port_ok = false;
  }
  else{
    printf("port open\n"); 
    open_port_ok = true;
  }
  struct termios options;

  //read current settings
  tcgetattr(serial, &options);
  //raw mode -> turn all the special features off
  //+ use 8N1
  cfmakeraw(&options);
  //set baudrate
  cfsetispeed(&options, BAUD);
  cfsetospeed(&options, BAUD);
  //enable receiving
  options.c_cflag |=CREAD;
  //minimum number of chars to read
  options.c_cc[VMIN]=22;
  //timeout reading
  options.c_cc[VTIME]=2;
  //no flowcontrol
  #ifdef CNEW_RTSCTS
	options.c_cflag &= ~CNEW_RTSCTS;
  #endif
  //write options
  tcsetattr(serial, TCSANOW, &options);

  sleep(3);
  count_packet_send = 0;
} 

// ----------- Main --------------------------------------------------------
int main(int argc, char **argv)
{

  ros::init(argc, argv, "mwSerial");
  
  ros::NodeHandle node_handle_;
  Combiner c(node_handle_);

  //qcore.msg import mav_rcdata
  ros::Subscriber sub_move = node_handle_.subscribe("qcore/rc", 1000, &Combiner::controlCallback, &c);

  ros::spin();

  closeSerial();

  return 0;
}
